Type QFolderTree Extends QTreeView
  PRIVATE:
    hImgLst       As Long
    ExpandedIndex As Long
    CurrentDir    As String
    DirList       As QFileListBox

  PUBLIC:
    SafeMode   As Byte
    InitialDir As String

  PRIVATE:
    Function GetVolInfo(ByVal path As String) As Long
        Dim RetVal                 As Long
        Dim SerialNumber           As Long
        Dim MaximumComponentLength As DWord
        Dim VolumeNameBuffer       As String
        Dim FileSystemNameBuffer   As String
        Dim RootPathName           As String

        RetVal                 = 0
        SerialNumber           = 0
        MaximumComponentLength = 0
        VolumeNameBuffer       = Space$(255)
        FileSystemNameBuffer   = Space$(255)
        RootPathName           = Space$(255)

        RootPathName = path
        RetVal = GetVolumeInformation(VarPtr(RootPathName), VarPtr(VolumeNameBuffer), 255, VarPtr(SerialNumber), VarPtr(MaximumComponentLength), 0, VarPtr(FileSystemNameBuffer), 255)

        Result = RetVal

        RetVal                 = 0
        SerialNumber           = 0
        MaximumComponentLength = 0
        VolumeNameBuffer       = ""
        FileSystemNameBuffer   = ""
        RootPathName           = ""
    End Function

    Function ExtractPathName(ByVal index As Long) As String
      Dim I    As Long
      Dim L    As Long
      Dim N    As Long
      Dim Path As String

      I    = index
      L    = This.Item(I).Level
      N    = 0
      Path = This.Item(I).Text

      Do
        N = This.Item(I).Level
        If N = (L - 1) Then
          Path = This.Item(I).Text & "\" & Path
          Dec(L, 1)
        End If
        Dec(I, 1)
      Loop Until L = 0

      Result = Path

      I    = 0
      L    = 0
      N    = 0
      Path = ""
    End Function

    Sub SetIcon(ByVal index As Long, ByVal path As String)
      Dim PathName As String

      PathName = ""

      Select Case GetDriveType(path)
        Case 2, 3, 4, 5, 6
          PathName = IIf(Right$(path,1) = "\", path, path & "\")
          SHGetFileInfo(PathName, 0, ftSFI, SizeOf(ftSFI), SHGFI_SMALLICON Or SHGFI_SYSICONINDEX)
          This.Item(index).ImageIndex = ftSFI.iIcon
          This.Item(index).SelectedIndex = ftSFI.iIcon
          DestroyIcon(ftSFI.hIcon)
        Case Else
          PathName = IIf(Right$(path,1) = "\", path - "\", path)
          SHGetFileInfo(PathName, 0, ftSFI, SizeOf(ftSFI), SHGFI_SMALLICON Or SHGFI_SYSICONINDEX)
          This.Item(index).ImageIndex = ftSFI.iIcon
          This.Item(index).SelectedIndex = ftSFI.iIcon
          DestroyIcon(ftSFI.hIcon)
      End Select

      PathName = ""
    End Sub

    Sub GetFakeSubFolders(ByVal index As Long, ByVal path As String)
      Dim I             As Long
      Dim N             As Long
      Dim RetVal        As Long
      Dim FullPathName  As String
      Dim SubFolderName As String
      Dim PathName      As String

      I             = 0
      N             = 0
      FullPathName  = ""
      SubFolderName = ""
      PathName      = IIf(Right$(path,1) = "\", path, path & "\")

      RetVal = IIf(This.Item(index).Level = 0, This.GetVolInfo(PathName), GetDriveType(path))

      If RetVal Then
        Select Case This.SafeMode
          Case False
            This.DirList.Mask = ("*.*")
            This.DirList.AddFileTypes(ftReadOnly, ftHidden, ftSystem, ftDirectory)
            This.DirList.DelFileTypes(ftNormal)
            This.DirList.Directory = PathName
            If This.DirList.ItemCount > 0 Then
              I = 0
              Do
                SubFolderName = This.DirList.Item(I) - "[" - "]"
                If Left$(SubFolderName,1) <> Chr$(46) Then
                  FullPathName = PathName & SubFolderName
                  If GetFileAttributes(FullPathName) And FILE_ATTRIBUTE_DIRECTORY Then Inc(N, 1)
                End If
                Inc(I, 1)
              Loop Until (N > 0) Or (I = This.DirList.ItemCount)
            End If
          Case True
            This.DirList.Mask = ("*.")
            This.DirList.AddFileTypes(ftDirectory)
            This.DirList.DelFileTypes(ftNormal)
            This.DirList.Directory = PathName
            If This.DirList.ItemCount > 0 Then
              I = 0
              Do
                SubFolderName = This.DirList.Item(I) - "[" - "]"
                If Left$(SubFolderName,1) <> Chr$(46) Then
                  FullPathName = PathName & SubFolderName
                  If GetFileAttributes(FullPathName) And FILE_ATTRIBUTE_DIRECTORY Then Inc(N, 1)
                End If
                Inc(I, 1)
              Loop Until (N > 0) Or (I = This.DirList.ItemCount)
            End If
        End Select
        If N > 0 Then This.AddChildItems(index, " ")
      End If

      I             = 0
      N             = 0
      RetVal        = 0
      FullPathName  = ""
      SubFolderName = ""
      PathName      = ""
    End Sub

    Sub GetRealSubFolders(ByVal index As Long, ByVal path As String)
      Dim I             As Long
      Dim ChildIndex    As Long
      Dim SubFolderName As String
      Dim FullPathName  As String
      Dim PathName      As String

      I             = 0
      ChildIndex    = 0
      SubFolderName = ""
      FullPathName  = ""
      PathName      = IIf(Right$(path,1) = "\", path, path & "\")

      Select Case This.SafeMode
        Case False
          This.DirList.Mask = ("*.*")
          This.DirList.AddFileTypes(ftReadOnly, ftHidden, ftSystem, ftDirectory)
          This.DirList.DelFileTypes(ftNormal)
          This.DirList.Directory = PathName
          For I = 0 To (This.DirList.ItemCount-1)
            SubFolderName = This.DirList.Item(I) - "[" - "]"
            If Left$(SubFolderName,1) <> Chr$(46) Then
              FullPathName = PathName & SubFolderName
              If GetFileAttributes(FullPathName) And FILE_ATTRIBUTE_DIRECTORY Then
                This.AddChildItems(index, SubFolderName)
                ChildIndex = index + This.Item(index).Count
                This.SetIcon(ChildIndex, FullPathName)
              End If
            End If
          Next I
        Case True
          This.DirList.Mask = ("*.")
          This.DirList.AddFileTypes(ftDirectory)
          This.DirList.DelFileTypes(ftNormal)
          This.DirList.Directory = PathName
          For I = 0 To (This.DirList.ItemCount-1)
            SubFolderName = This.DirList.Item(I) - "[" - "]"
            If Left$(SubFolderName,1) <> Chr$(46) Then
              FullPathName = PathName & SubFolderName
              This.AddChildItems(index, SubFolderName)
              ChildIndex = index + This.Item(index).Count
              This.SetIcon(ChildIndex, FullPathName)
            End If
          Next I
      End Select

      I             = 0
      ChildIndex    = 0
      SubFolderName = ""
      FullPathName  = ""
      PathName      = ""
    End Sub

    Sub FolderTree_OnExpanding(ByVal index As Long, ByVal allowed As Long)
      Dim I As Long
      Dim N As Long

      I = 0
      N = 0

      If This.ExpandedIndex <> index Then
        This.Cursor = crHourGlass
        This.Enabled = False
        This.DelItems(index + 1)
        This.GetRealSubFolders(index, This.ExtractPathName(index))
        N = index + This.Item(index).Count
        For I = N To (index + 1) Step -1
          This.GetFakeSubFolders(I, This.ExtractPathName(I))
        Next I
        This.Enabled = True
        This.Cursor = crDefault
        This.ExpandedIndex = index
        SetFocus(This.Handle)
      End If

      I = 0
      N = 0
    End Sub

    Sub FolderTree_OnCollapsing(ByVal index As Long, ByVal allowed As Long)
      Dim ExitFlag As Byte
      Dim C        As Long
      Dim I        As Long
      Dim L        As Long
      Dim N        As Long

      ExitFlag = 0
      C        = 0
      I        = 0
      L        = 0
      N        = 0

      I = index + 1
      L = This.Item(I).Level
      Do
        C = 1
        If I = (This.ItemCount-1) Then
          ExitFlag = 1
        ElseIf This.Item(I).Level >= L Then
          Inc(N, 1)
          If This.Item(I).HasChildren = True Then
            Inc(N, This.Item(I).Count)
            Inc(C, This.Item(I).Count)
          End If
        ElseIf This.Item(I).Level < L Then
          ExitFlag = 1
        End If
        Inc(I, C)
      Loop Until ExitFlag = 1
      C = index + N
      For I = C To (index + 1) Step -1
        This.DelItems(I)
      Next I
      This.ExpandedIndex = -1
      This.AddChildItems(index, " ")

      ExitFlag = 0
      C        = 0
      I        = 0
      L        = 0
      N        = 0
    End Sub

    Sub SetInitialDir(ByVal path As String)
      Dim I             As Long
      Dim N             As Long
      Dim PathName      As String
      Dim FolderName    As String
      Dim ParsedFolders As QStringList

      I          = 0
      N          = 0
      PathName   = IIf(Right$(path,1) = "\", UCase$(path), UCase$(path & "\"))
      FolderName = ""
      ParsedFolders.Clear
      ParsedFolders.Duplicates = dupAccept

      If DirExists(PathName) Then
        ParsedFolders.Parse(PathName, "\")
        Do
          FolderName = UCase$(This.Item(I).Text)
          If FolderName = UCase$(ParsedFolders.Item(N)) Then
            This.ItemIndex = I
            This.Expand(I, False)
            Inc(N, 1)
          End If
          Inc(I, 1)
        Loop Until N = ParsedFolders.ItemCount
        This.ItemIndex = (I - 1)
        This.CurrentDir = PathName
      Else
        This.InitialDir = Command$(0) - Application.ExeName
        PathName = IIf(Right$(This.InitialDir,1) = "\", UCase$(This.InitialDir), UCase$(This.InitialDir & "\"))
        ParsedFolders.Parse(PathName, "\")
        Do
          FolderName = UCase$(This.Item(I).Text)
          If FolderName = UCase$(ParsedFolders.Item(N)) Then
            This.ItemIndex = I
            This.Expand(I, False)
            Inc(N, 1)
          End If
          Inc(I, 1)
        Loop Until N = ParsedFolders.ItemCount
        This.ItemIndex = (I - 1)
        This.CurrentDir = PathName
      End If

      I          = 0
      N          = 0
      PathName   = ""
      FolderName = ""
      ParsedFolders.Clear
    End Sub

  PUBLIC:
    Function Directory() As String
      Result = This.CurrentDir
    End Function

    Sub FolderTree_OnClick
      Dim FullPathName As String

      FullPathName = ""

      If This.Item(This.ItemIndex).Level > 0 Then
        If This.Item(This.ItemIndex).Selected Then
          FullPathName = This.ExtractPathName(This.ItemIndex) & "\"
          If FullPathName <> This.CurrentDir Then
            This.CurrentDir = FullPathName
          End If
        End If
      ElseIf This.Item(This.ItemIndex).Level = 0 Then
        If This.Item(This.ItemIndex).Selected Then
          FullPathName = This.Item(This.ItemIndex).Text & "\"
          If FullPathName <> This.CurrentDir Then
            This.CurrentDir = FullPathName
          End If
        End If
      End If

      FullPathName = ""
    End Sub

    Sub FolderTree_OnKeyUp(ByVal key As Byte, ByVal shift As Long)
      This.FolderTree_OnClick
    End Sub

    Sub Initialize(ByRef form As QForm)
      Dim I           As Long
      Dim RetVal      As Long
      Dim DriveLetter As String

      I           = 0
      RetVal      = 0
      DriveLetter = ""

      This.DirList.Parent = form

      This.hImgLst = SHGetFileInfo("c:\", 0, ftSFI, SizeOf(ftSFI), SHGFI_SMALLICON Or SHGFI_SYSICONINDEX)
      SendMessage(This.Handle, TVM_SETIMAGELIST, TVSIL_NORMAL, This.hImgLst)
      SendMessage(This.Handle, TVM_SETIMAGELIST, TVSIL_STATE, This.hImgLst)

      RetVal = GetLogicalDrives()

      For I = 0 To 25
        If (RetVal And 2 ^ I) <> 0 Then
          DriveLetter = Chr$(65 + I) & ":"
          This.AddItems(DriveLetter)
          This.SetIcon(This.ItemCount-1, DriveLetter)
          This.GetFakeSubFolders(This.ItemCount-1, DriveLetter)
        End If
      Next I

      This.SetInitialDir(This.InitialDir)

      hFTSItem = SendMessageAPI(This.Handle, TVM_GETNEXTITEM, TVGN_CARET, 0)
      SendMessage(This.Handle, TVM_EXPAND, TVE_EXPAND, hFTSItem)

      I           = 0
      RetVal      = 0
      DriveLetter = ""
    End Sub

    Constructor
      hImgLst         = 0
      ReadOnly        = True
      OnExpanding     = This.FolderTree_OnExpanding
      OnCollapsing    = This.FolderTree_OnCollapsing
      DirList.Visible = False
      ExpandedIndex   = -1
      CurrentDir      = ""

      SafeMode    = True
      InitialDir  = Command$(0) - Application.ExeName
      Left        = 0
      Top         = 0
      Width       = 320
      Height      = 240
      BorderStyle = bsNone
      Font.Name   = "Arial"
      Font.Size   = 9
    End Constructor
End Type
