Type QPrint Extends QObject
PRIVATE:
    DM              As DEVICEMODEA_UDT      'Used for copying the DEVMODE information from memory
                                            ' (can be used to find out what changes have been made from the default information)
    PSD             As PAGESETUPDIALOG_UDT  'Used for making changes to the defaults prior to printing and previewing
    PD              As PRINTDIALOG_UDT      'Used for making changes to the defaults prior to printing
    DI              As DOCINFO_UDT          'Used as part of the printing process
    FR              As FORMATRANGE_UDT      'Used as part of the printing process and previewing process
    GTL             As GETTEXTLENGTHEX_UDT  'Used as part of the printing process and previewing process
    WR              As RECT_UDT             'Used as part of the previewing process (for the dimensions of the preview Form)
    rcBound         As RECT_UDT             'Used as part of the previewing process (for the creation of an EMF image)
    rcParentSize    As RECT_UDT             'Used as part of the previewing process (for the dimensions of the parent Panel)
    rcChildSize     As RECT_UDT             'Used as part of the previewing process (for the dimensions of the child Panel)
    rcChildMargins  As RECT_UDT             'Used as part of the previewing process (for the displaying of user margins on the child Panel)
    rcParentMargins As RECT_UDT             'Used as part of the previewing process (for the displaying of user margins on the parent Panel)
    rcPrint         As RECT_UDT             'Used as part of the both the previewing process and the printing process

    mhDevMode As Long  'Variable that holds a "handle" to a DevModeA structure that's stored in memory.
                       ' (note: a "handle" is not the same as a "pointer")
                       ' (this handle will be used to find out if the .GetDevModeDefaults subroutine was called first)
    mpDevMode As Long  'Variable that holds a "pointer" to the "handle" of a DevModeA structure.
                       ' The .StoreMemoryDevMode subroutine will use this pointer to store the memory DevModeA structure into the accessible .DM structure.
                       ' (the memory DevModeA structure was automatically generated and stored in memory by the .GetDevModeDefaults subroutine)
    UserUnits  As Long  'Variable that holds a value that can be used to determine if the user's computer is set up to use inches or millimeters.
    hWndForm   As Long  'Variable that holds the handle of the main application's form.
    hWndEdit   As Long  'Variable that holds the handle of the main application's richedit.
    PageIndex  As Long  'Variable that holds the page index that is currently being displayed in the child Panel's Device Context.
    hdcMeta    As Long  'Variable that holds the Device Context of an Enhanced Meta File image.
    hdcPanel   As Long  'Variable that holds the Device Context of a child Panel used for displaying an EMF image.
    hEMF       As Long  'Variable that holds the handle of an Enhanced Meta File image.

    PreviewForm   As QForm       'Window for displaying a preview.
    CtrlPanel     As QPanel      'Container for holding the First, Previous, Next, and Last buttons.
    CtrlFirst     As QButton     'Button for jumping to the First page (of documents with multiple pages).
    CtrlPrevious  As QButton     'Button for viewing the Previous page (of documents with multiple pages).
    CtrlNext      As QButton     'Button for viewing the Next page (of documents with multiple pages).
    CtrlLast      As QButton     'Button for jumping to the Last page (of documents with multiple pages).
    CtrlPrint     As QButton     'Button for printing the preview page that is currently being viewed.
    ViewScrollbox As QScrollBox  'Container for holding the components that are used for displaying a preview page (allows for scrolling).
    ParentPanel   As QPanel      'Component that will be used to give the preview a slighty truer represetation of the margins
    ParentCanvas  As QCanvas     'Component that will be used for displaying the user margins on the parent Panel
    ChildPanel    As QPanel      'Component that will be used for displaying an EMF image "on" (think of it as a... movie screen)
                                  ' (the EMF image represents a document page as it would look if printed).
    ChildCanvas   As QCanvas     'Component that will be used for displaying the user margins on the child Panel
                                 ' (also... it's OnPaint event will be used to always correctly display the EMF image when the user makes use of scrollbars)

    Page As QStringList  'The Page stringlist has a dual purpose.
                         ' a) the ItemCount will indicate how many pages a document has for a selected printer.
                         ' b) each Item in the list will hold an index of the first character of each page that can be printed.
                         ' They are both calculated in the .CalculatePageCount subroutine.
                         ' They are both used when previewing a document.
                         ' They are also both used in the .PrtDialog subroutine (to print an entire document or just a single range of pages from a document).
                         ' If you were to port this code to another programming language... you'll probably want to use an array.

PUBLIC:
    DocName As String  'Variable that is a null-terminated string (required for printing).
                       ' If the .PrtDialog subroutine is called and the string contains nothing...
                       ' a generic document name will be assigned.

PRIVATE:
    'This function determines if the computer is setup to use inches or millimeters.
    Function GetLocaleInfo_iMeasure () As Long
        Dim sBuffer As String
        Dim iRet    As Long

        sBuffer = String$(2, 32)
        iRet = GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IMEASURE, VarPtr(sBuffer), 2)

        Result = Val(Mid$(sBuffer,1,1)) '0 = millimeters, !0 =  inches
    End Function

    'This subroutine stores the DeviceModeA structure that's in memory into an accessible structure.
    ' (so that you can check for changes, to the defaults, that a user may have made)
    Sub StoreMemoryDevMode (ByVal lpDevMode As Long)
        MemCpy(This.DM, lpDevMode, SizeOf(This.DM))
    End Sub

    'This subroutine gathers default information from the registry and stores it in...
    ' the PageSetupDialog structure,
    ' a DeviceModeA structure (that'll automatically be generated in memory),
    ' and in some global variables.
    'After the DeviceModeA structure is generated in memory...
    ' it will be stored in an accessible structure by the .StoreMemoryDevMode subroutine.
    'The information gathered here will be used for printing and/or previewing.
    Sub GetDevModeDefaults ()
        MemSet(This.PSD, 0, SizeOf(This.PSD))

        This.PSD.lStructSize = SizeOf(This.PSD)

        If This.UserUnits = 0 Then
            This.PSD.Flags = PSD_RETURNDEFAULT Or PSD_INHUNDREDTHSOFMILLIMETERS
        Else
            This.PSD.Flags = PSD_RETURNDEFAULT Or PSD_INTHOUSANDTHSOFINCHES
        End If

        'Calling PageSetupDlg with the above flags tells it not to display a dialog box
        ' and to only read defaults from the registry and generate a DevModeA structure in memory.
        ' (footnote: it reads a DevModeW from the registry, but generates a DevModeA in memory)
        PageSetupDlg(This.PSD)

        This.mhDevMode   = This.PSD.hDevMode
        This.mpDevMode = GlobalLock(This.mhDevMode)
        GlobalUnLock(This.mhDevMode)

        This.StoreMemoryDevMode(This.mpDevMode)
    End Sub

    'This subroutine calculates the width and height... that an EMF image should be (in millimeters).
    ' (footnote: that's what printers do. they print an "EMF image" of whatever it is... you are trying to print)
    ' (          why i didn't think of this workaround on my own is beyond me. especially considering i created a .spl file viewer)
    'Anyways... it also calculates...
    ' the width and height that the Panel should be (in screen pixels).
    ' the rectangle that will be used to center the the EMF image onto the Panel (in twips).
    ' the rectangle that will be used to center the the EMF image onto the Printer (in twips).
    'The calculations must take (all of the following) into consideration...
    ' (note: not in any specific order)
    ' a) the user's unit of measure
    ' b) is it portrait or landscape?
    ' c) the size of the (default or selected) paper/envolope/card/????
    ' d) the unprintable areas of the (default or selected) paper/envolope/card/????
    ' e) the (default or selected) margins
    'All dimensions are gathered/calculated as "printer DPIs"... then converted to what is appropriate for their usage.
    'After the calculations are done... they need to be stored in (global) rectangle structures.
    Sub CalculateDimensions ()
        Dim hDesktop   As Long
        Dim hdcDesktop As Long

        Dim lPrntDPIx As Long
        Dim lPrntDPIy As Long

        Dim lScrnDPIx As Long
        Dim lScrnDPIy As Long

        Dim lPrntWidth  As Long
        Dim lPrntHeight As Long
        Dim lPageWidth  As Long
        Dim lPageHeight As Long

        Dim Offset     As RECT_UDT
        Dim UserMargin As RECT_UDT
        Dim TrueMargin As RECT_UDT

        'Need to get the screen's Device Context to get the screen's DPIs
        hDesktop   = GetDesktopWindow()
        hdcDesktop = GetWindowDC(hDesktop)
        lScrnDPIx  = GetDeviceCaps(hdcDesktop, LOGPIXELSX)
        lScrnDPIy  = GetDeviceCaps(hdcDesktop, LOGPIXELSY)
        ReleaseDC(hDesktop, hdcDesktop)

        'Need to get the printer's Device Context to get the printer's DPIs and unprintable areas
        MemSet(This.PD, 0, SizeOf(This.PD))
        This.PD.lStructSize = SizeOf(This.PD)
        This.PD.Flags       = PD_RETURNDC Or PD_RETURNDEFAULT
        'Calling PrintDlg with the above flags tells it to not display a dialog box
        ' and to fill the PD structure with defaults.
        PrintDlg(This.PD)
        lPrntDPIx     = GetDeviceCaps(This.PD.hDC, LOGPIXELSX)
        lPrntDPIy     = GetDeviceCaps(This.PD.hDC, LOGPIXELSY)
        Offset.Left   = GetDeviceCaps(This.PD.hDC, PHYSICALOFFSETX)
        Offset.Top    = GetDeviceCaps(This.PD.hDC, PHYSICALOFFSETY)
        Offset.Right  = GetDeviceCaps(This.PD.hDC, PHYSICALWIDTH)  - GetDeviceCaps(This.PD.hDC, HORZRES) - GetDeviceCaps(This.PD.hDC, PHYSICALOFFSETX)
        Offset.Bottom = GetDeviceCaps(This.PD.hDC, PHYSICALHEIGHT) - GetDeviceCaps(This.PD.hDC, VERTRES) - GetDeviceCaps(This.PD.hDC, PHYSICALOFFSETY)
        DeleteDC(This.PD.hDC)

        'The (default or selected) user margins
        If This.UserUnits = 0 Then
            UserMargin.Left   = Round((This.PSD.rtMarginLeft   / 2540) * lPrntDPIx)
            UserMargin.Top    = Round((This.PSD.rtMarginTop    / 2540) * lPrntDPIy)
            UserMargin.Right  = Round((This.PSD.rtMarginRight  / 2540) * lPrntDPIx)
            UserMargin.Bottom = Round((This.PSD.rtMarginBottom / 2540) * lPrntDPIy)
        Else
            UserMargin.Left   = Round((This.PSD.rtMarginLeft   / 1000) * lPrntDPIx)
            UserMargin.Top    = Round((This.PSD.rtMarginTop    / 1000) * lPrntDPIy)
            UserMargin.Right  = Round((This.PSD.rtMarginRight  / 1000) * lPrntDPIx)
            UserMargin.Bottom = Round((This.PSD.rtMarginBottom / 1000) * lPrntDPIy)
        End If

        'The width and height of the paper/envolope/card/????
        If This.UserUnits = 0 Then
            lPageWidth  = Round((This.PSD.ptPaperSizeX / 2540) * lPrntDPIx)
            lPageHeight = Round((This.PSD.ptPaperSizeY / 2540) * lPrntDPIy)
        Else
            lPageWidth  = Round((This.PSD.ptPaperSizeX / 1000) * lPrntDPIx)
            lPageHeight = Round((This.PSD.ptPaperSizeY / 1000) * lPrntDPIy)
        End If

        'The width and height of the printable area (of the paper/envolope/card/????)
        If This.DM.dmOrientation = DMORIENT_LANDSCAPE Then
            lPrntWidth  = (lPageWidth  - (Offset.Top  + Offset.Bottom))
            lPrntHeight = (lPageHeight - (Offset.Left + Offset.Right))
        Else
            lPrntWidth  = (lPageWidth  - (Offset.Left + Offset.Right))
            lPrntHeight = (lPageHeight - (Offset.Top  + Offset.Bottom))
        End If

        'Calculate the difference between the unprintable areas and the user margins
        If This.DM.dmOrientation = DMORIENT_LANDSCAPE Then
            TrueMargin.Left   = UserMargin.Left   - Offset.Top
            TrueMargin.Top    = UserMargin.Top    - Offset.Right
            TrueMargin.Right  = UserMargin.Right  - Offset.Bottom
            TrueMargin.Bottom = UserMargin.Bottom - Offset.Left
        Else
            TrueMargin.Left   = UserMargin.Left   - Offset.Left
            TrueMargin.Top    = UserMargin.Top    - Offset.Top
            TrueMargin.Right  = UserMargin.Right  - Offset.Right
            TrueMargin.Bottom = UserMargin.Bottom - Offset.Bottom
        End If

        'Clear the "global" rectangles
        MemSet(This.rcBound,        0, SizeOf(This.rcBound))
        MemSet(This.rcChildSize,    0, SizeOf(This.rcChildSize))
        MemSet(This.rcParentSize,   0, SizeOf(This.rcParentSize))
        MemSet(This.rcChildMargins, 0, SizeOf(This.rcChildMargins))
        MemSet(This.rcPrint,        0, SizeOf(This.rcPrint))

        'Now make the final calculations and convert them to what is appropriate for their usage.
        This.rcBound.Right  = Fix((lPrntWidth  / lPrntDPIx) * 25.4) * 100
        This.rcBound.Bottom = Fix((lPrntHeight / lPrntDPIy) * 25.4) * 100

        If This.DM.dmOrientation = DMORIENT_LANDSCAPE Then
            This.rcChildSize.Left = Round((Offset.Top   / lPrntDPIx) * lScrnDPIx)
            This.rcChildSize.Top  = Round((Offset.Right / lPrntDPIy) * lScrnDPIy)
        Else
            This.rcChildSize.Left = Round((Offset.Top / lPrntDPIx) * lScrnDPIx)
        End If
        This.rcChildSize.Right  = Round((lPageWidth  / lPrntDPIx) * lScrnDPIx)
        This.rcChildSize.Bottom = Round((lPageHeight / lPrntDPIy) * lScrnDPIy)

        If This.DM.dmOrientation = DMORIENT_LANDSCAPE Then
            This.rcParentSize.Right  = Round((Offset.Top  / lPrntDPIx) * lScrnDPIx) + This.rcChildSize.Right  + Round((Offset.Bottom / lPrntDPIx) * lScrnDPIx)
            This.rcParentSize.Bottom = Round((Offset.Left / lPrntDPIy) * lScrnDPIy) + This.rcChildSize.Bottom + Round((Offset.Right  / lPrntDPIy) * lScrnDPIy)
        Else
            This.rcParentSize.Right  = Round((Offset.Left / lPrntDPIx) * lScrnDPIx) + This.rcChildSize.Right  + Round((Offset.Right  / lPrntDPIx) * lScrnDPIx)
            This.rcParentSize.Bottom = Round((Offset.Top  / lPrntDPIy) * lScrnDPIy) + This.rcChildSize.Bottom + Round((Offset.Bottom / lPrntDPIy) * lScrnDPIy)
        End If

        This.rcChildMargins.Left    = Round((TrueMargin.Left / lPrntDPIx) * lScrnDPIx) + This.rcChildSize.Left
        This.rcChildMargins.Top     = Round((TrueMargin.Top  / lPrntDPIy) * lScrnDPIx) + This.rcChildSize.Top
        This.rcChildMargins.Right   = Round(((lPageWidth  - TrueMargin.Right)  / lPrntDPIx) * lScrnDPIx)
        This.rcChildMargins.Bottom  = Round(((lPageHeight - TrueMargin.Bottom) / lPrntDPIy) * lScrnDPIy)

        This.rcParentMargins.Left    = This.rcChildMargins.Left   + This.rcChildSize.Left
        This.rcParentMargins.Top     = This.rcChildMargins.Top    + This.rcChildSize.Top
        This.rcParentMargins.Right   = This.rcChildMargins.Right  + This.rcChildSize.Left
        This.rcParentMargins.Bottom  = This.rcChildMargins.Bottom + This.rcChildSize.Top

        This.rcPrint.Left     = Round((TrueMargin.Left / lPrntDPIx) * 1440)
        This.rcPrint.Top      = Round((TrueMargin.Top  / lPrntDPIy) * 1440)
        This.rcPrint.Right    = Round(((lPrntWidth  - TrueMargin.Right)  / lPrntDPIx) * 1440)
        This.rcPrint.Bottom   = Round(((lPrntHeight - TrueMargin.Bottom) / lPrntDPIy) * 1440)
    End Sub

    'This subroutine counts how many pages it will take to print a documemt.
    ' (based on the dimensions that were calculated in the .CalculateDimensions subroutine)
    Sub CalculatePageCount (ByVal hEdit As Long)
        Dim iTextOut  As Long
        Dim iTextAmt  As Long
        Dim Classname As String
        Dim lRet      As Long

        iTextOut = 0
        This.Page.Clear
        This.Page.AddItems("0")

        This.GTL.flags = GTL_NUMCHARS
        This.GTL.codepage = CP_ACP

        Classname = Space$(255)
        GetClassName(hEdit, VarPtr(Classname), 255)
        If UCase$(Left$(Classname, Instr(Classname, Chr$(0))-1)) = "TRICHEDITS" Then
            iTextAmt = SendMessageAPI(hEdit, WM_GETTEXTLENGTH, 0, 0)
        Else
            iTextAmt = SendMessageAPI(hEdit, WM_GETTEXTLENGTHEX, This.GTL, 0)
        End If

        If iTextAmt Then
            This.Page.AddItems(STR$(iTextAmt + 1))
        Else
            This.Page.AddItems("-1")
        End If

        MemSet(This.PD, 0, SizeOf(This.PD))
        This.PD.lStructSize = SizeOf(This.PD)
        This.PD.Flags       = PD_RETURNDC Or PD_RETURNDEFAULT
        'Calling PrintDlg with the above flags tells it to not display a dialog box
        ' and to fill the PD structure with defaults.
        PrintDlg(This.PD)

        SendMessage(hEdit, EM_FORMATRANGE, 0, 0)

        While (iTextOut < iTextAmt)
            This.hdcMeta = CreateEnhMetaFile(This.PD.hDC, 0, This.rcBound, 0)
            If This.hdcMeta Then
                MemSet(This.FR, 0, SizeOf(This.FR))
                This.FR.hdc           = This.hdcMeta
                This.FR.hdcTarget     = This.hdcMeta
                This.FR.chrg_cpMin    = iTextOut
                This.FR.chrg_cpMax    = -1
                This.FR.rcPage_Left   = This.rcPrint.Left
                This.FR.rcPage_Top    = This.rcPrint.Top
                This.FR.rcPage_Right  = This.rcPrint.Right
                This.FR.rcPage_Bottom = This.rcPrint.Bottom
                This.FR.rc_Left       = This.FR.rcPage_Left
                This.FR.rc_Top        = This.FR.rcPage_Top
                This.FR.rc_Right      = This.FR.rcPage_Right
                This.FR.rc_Bottom     = This.FR.rcPage_Bottom

                iTextOut = SendMessageAPI(hEdit, EM_FORMATRANGE, 0, This.FR)

                If (iTextOut < iTextAmt) Then
                    This.Page.InsertItem(This.Page.ItemCount-1, STR$(iTextOut))
                    This.FR.rc_Bottom = This.FR.rcPage_Bottom
                End If

                lRet = CloseEnhMetaFile(This.hdcMeta)

                DeleteEnhMetaFile(lRet)
                DeleteDC(This.hdcMeta)
            Else
                iTextOut = iTextAmt + 1
            End If
        Wend

        SendMessage(hEdit, EM_FORMATRANGE, 0, 0)
        DeleteDC(This.PD.hDC)
    End Sub

    'This subroutine creates an EMF image (in memory) of a single page from the document.
    ' (based on the dimensions that were calculated in the .CalculateDimensions subroutine)
    Sub GeneratePage (ByVal hEdit As Long, ByVal index As Long)
        If This.hdcMeta Then DeleteDC(This.hdcMeta)

        MemSet(This.PD, 0, SizeOf(This.PD))
        This.PD.lStructSize = SizeOf(This.PD)
        This.PD.Flags       = PD_RETURNDC Or PD_RETURNDEFAULT
        'Calling PrintDlg with the above flags tells it to not display a dialog box
        ' and to fill the PD structure with defaults.
        PrintDlg(This.PD)

        This.hdcMeta = CreateEnhMetaFile(This.PD.hDC, 0, This.rcBound, 0)

        DeleteDC(This.PD.hDC)

        If This.hdcMeta Then
            MemSet(This.FR, 0, SizeOf(This.FR))
            This.FR.hdc           = This.hdcMeta
            This.FR.hdcTarget     = This.hdcMeta
            This.FR.chrg_cpMin    = Val(This.Page.Item(index))
            This.FR.chrg_cpMax    = Val(This.Page.Item(index+1))
            This.FR.rcPage_Left   = This.rcPrint.Left
            This.FR.rcPage_Top    = This.rcPrint.Top
            This.FR.rcPage_Right  = This.rcPrint.Right
            This.FR.rcPage_Bottom = This.rcPrint.Bottom
            This.FR.rc_Left       = This.FR.rcPage_Left
            This.FR.rc_Top        = This.FR.rcPage_Top
            This.FR.rc_Right      = This.FR.rcPage_Right
            This.FR.rc_Bottom     = This.FR.rcPage_Bottom

            SendMessage(hEdit, EM_FORMATRANGE, 0, 0)
            SendMessage(hEdit, EM_FORMATRANGE, 1, This.FR)
            SendMessage(hEdit, EM_FORMATRANGE, 1, 0)

            This.hEMF = CloseEnhMetaFile(This.hdcMeta)
        Else
            This.hEMF = 0
        End If
    End Sub

    'This subroutine first erases the Device Context of the child panel.
    ' It then displays an EMF image on that Device Context that represents a document page.
    ' (the Canvas OnPaint event... provides the perfect solution for displaying the EMF correctly if the user makes use of scrollbars)
    ' Lastly... ot draws 4 solid lines that represent the user margins.
    ' (it probably would look better with dash lines, but I kept it simple)
    ' The location of the 4 lines were calculated in the .CalculateDimensions subroutine.
    Sub ChildCanvas_OnPaint ()
        If (This.hdcPanel = 0) And (This.hEMF = 0) Then
            This.ChildCanvas.FillRect(0, 0, This.ChildCanvas.Width, This.ChildCanvas.Height, RGB(205,205,205))
        Else
            This.ChildCanvas.FillRect(0, 0, This.ChildCanvas.Width, This.ChildCanvas.Height, RGB(255,255,255))
            PlayEnhMetaFile(This.hdcPanel, This.hEMF, This.rcChildSize)
            This.ChildCanvas.Line(This.rcChildMargins.Left, 0, This.rcChildMargins.Left, This.ChildCanvas.Height, RGB(205,205,205))
            This.ChildCanvas.Line(0, This.rcChildMargins.Top, This.ChildCanvas.Width, This.rcChildMargins.Top, RGB(205,205,205))
            This.ChildCanvas.Line(This.rcChildMargins.Right, 0, This.rcChildMargins.Right, This.ChildCanvas.Height, RGB(205,205,205))
            This.ChildCanvas.Line(0, This.rcChildMargins.Bottom, This.ChildCanvas.Width, This.rcChildMargins.Bottom, RGB(205,205,205))
        End If
    End Sub

    'This subroutine simply draws 4 solid lines that represent the user margins.
    ' (it probably would look better with dash lines, but I kept it simple)
    ' The location of the 4 lines were calculated in the .CalculateDimensions subroutine.
    ' The margins drawn on the child canvas needed to be extended and this seemed like the simplest solution.
    Sub ParentCanvas_OnPaint
        This.ParentCanvas.FillRect(0, 0, This.ParentCanvas.Width, This.ParentCanvas.Height, RGB(255,255,255))
        This.ParentCanvas.Line(This.rcParentMargins.Left, 0, This.rcParentMargins.Left, This.ParentCanvas.Height, RGB(205,205,205))
        This.ParentCanvas.Line(0, This.rcParentMargins.Top, This.ParentCanvas.Width, This.rcParentMargins.Top, RGB(205,205,205))
        This.ParentCanvas.Line(This.rcParentMargins.Right, 0, This.rcParentMargins.Right, This.ParentCanvas.Height, RGB(205,205,205))
        This.ParentCanvas.Line(0, This.rcParentMargins.Bottom, This.ParentCanvas.Width, This.rcParentMargins.Bottom, RGB(205,205,205))
    End Sub

    'This subroutine minimizes the main application's form,
    ' then moves the preview form to the foreground.
    Sub PreviewForm_OnShow ()
        ShowWindow(This.hWndForm, 6)
        SetForegroundWindow(This.PreviewForm.Handle)
    End Sub

    'This subroutine cleans up the mess created by previewing,
    ' then restores the main application's form.
    Sub PreviewForm_OnClose ()
        This.ViewScrollbox.HorzPosition = 0 'Oddity... you have to reset the Scrollox's scrollbars back to zero,
        This.ViewScrollbox.VertPosition = 0 'otherwise it screws up the preview the next time it's called.
        If This.hdcPanel Then ReleaseDC(This.ChildPanel.Handle, This.hdcPanel)
        This.hdcPanel = 0
        If This.hEMF Then DeleteEnhMetaFile(This.hEMF)
        This.hEMF = 0
        If This.hdcMeta Then DeleteDC(This.hdcMeta)
        This.hdcMeta = 0
        ShowWindow(This.hWndForm, 9)
    End Sub

    'This subroutine calls the .GeneratePage subroutine (with the index of the first page)
    ' and the .ChildCanvas_OnPaint subroutine
    ' to display the first page (if possible) of a document with multiple pages,
    ' then updates components to inform the user of what they are seeing and what they can do.
    Sub CtrlFirst_OnClick ()
        If This.PageIndex > 0 Then
            This.PageIndex = 0
            If This.hEMF Then DeleteEnhMetaFile(This.hEMF)
            This.GeneratePage(This.hWndEdit, This.PageIndex)
            This.ChildCanvas_OnPaint
        End If
        This.CtrlPanel.Caption    = "Page: " + STR$(This.PageIndex + 1) + " of " + STR$(This.Page.ItemCount - 1)
        This.CtrlFirst.Enabled    = IIf(This.PageIndex > 1, True, False)
        This.CtrlPrevious.Enabled = IIf(This.PageIndex > 0, True, False)
        This.CtrlNext.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 2), True, False)
        This.CtrlLast.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 3), True, False)
    End Sub

    'This subroutine calls the .GeneratePage subroutine (with a decremented page index)
    ' and the .ChildCanvas_OnPaint subroutine
    ' to display the previous page (if possible) of a document with multiple pages,
    ' then updates components to inform the user of what they are seeing and what they can do.
    Sub CtrlPrevious_OnClick ()
        If This.PageIndex > 0 Then
            This.PageIndex = This.PageIndex - 1
            If This.hEMF Then DeleteEnhMetaFile(This.hEMF)
            This.GeneratePage(This.hWndEdit, This.PageIndex)
            This.ChildCanvas_OnPaint
        End If
        This.CtrlPanel.Caption    = "Page: " + STR$(This.PageIndex + 1) + " of " + STR$(This.Page.ItemCount - 1)
        This.CtrlFirst.Enabled    = IIf(This.PageIndex > 1, True, False)
        This.CtrlPrevious.Enabled = IIf(This.PageIndex > 0, True, False)
        This.CtrlNext.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 2), True, False)
        This.CtrlLast.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 3), True, False)
    End Sub

    'This subroutine calls the .GeneratePage subroutine (with an incremented page index)
    ' and the .ChildCanvas_OnPaint subroutine
    ' to display the next page (if possible) of a document with multiple pages,
    ' then updates components to inform the user of what they are seeing and what they can do.
    Sub CtrlNext_OnClick ()
        If This.PageIndex < (This.Page.ItemCount - 2) Then
            This.PageIndex = This.PageIndex + 1
            If This.hEMF Then DeleteEnhMetaFile(This.hEMF)
            This.GeneratePage(This.hWndEdit, This.PageIndex)
            This.ChildCanvas_OnPaint
        End If
        This.CtrlPanel.Caption    = "Page: " + STR$(This.PageIndex + 1) + " of " + STR$(This.Page.ItemCount - 1)
        This.CtrlFirst.Enabled    = IIf(This.PageIndex > 1, True, False)
        This.CtrlPrevious.Enabled = IIf(This.PageIndex > 0, True, False)
        This.CtrlNext.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 2), True, False)
        This.CtrlLast.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 3), True, False)
    End Sub

    'This subroutine calls the .GeneratePage subroutine (with the index of the last page)
    ' and the .ChildCanvas_OnPaint subroutine
    ' to display the last page (if possible) of a document with multiple pages,
    ' then updates components to inform the user of what they are seeing and what they can do.
    Sub CtrlLast_OnClick ()
        If This.PageIndex < (This.Page.ItemCount - 1) Then
            This.PageIndex = (This.Page.ItemCount - 2)
            If This.hEMF Then DeleteEnhMetaFile(This.hEMF)
            This.GeneratePage(This.hWndEdit, This.PageIndex)
            This.ChildCanvas_OnPaint
        End If
        This.CtrlPanel.Caption    = "Page: " + STR$(This.PageIndex + 1) + " of " + STR$(This.Page.ItemCount - 1)
        This.CtrlFirst.Enabled    = IIf(This.PageIndex > 1, True, False)
        This.CtrlPrevious.Enabled = IIf(This.PageIndex > 0, True, False)
        This.CtrlNext.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 2), True, False)
        This.CtrlLast.Enabled     = IIf(This.PageIndex < (This.Page.ItemCount - 3), True, False)
    End Sub

    'This subroutine prints the currently being viewed document page.
    ' All it's doing is replacing two of the fields in the FormatRange structure
    ' with values that are appropriate for printing (the printer's DC)...
    ' doing a single page print...
    ' then restoring the altered fields with the values that were appropriate for previewing (the panel's DC).
    Sub CtrlPrint_OnClick ()
        Dim szString  As String * 200

        MemSet(This.PD, 0, SizeOf(This.PD))
        This.PD.lStructSize = SizeOf(This.PD)
        This.PD.Flags = PD_RETURNDC Or PD_RETURNDEFAULT
        'Calling PrintDlg with the above flags tells it to not display a dialog box
        ' and to fill the PD structure with defaults.
        PrintDlg(This.PD)

        This.FR.hdc       = This.PD.hDC
        This.FR.hdcTarget = This.PD.hDC

        If This.DocName = "" Then
            szString = "QPrint: NoName"
        Else
            szString = "QPrint: " & Mid$(This.DocName, rInstr(This.DocName, "\") + 1)
        End If

        MemSet(This.DI, 0, SizeOf(This.DI))
        This.DI.cbSize      = SizeOf(This.DI)
        This.DI.lpszDocName = VarPtr(szString)
        This.DI.lpszOutput  = 0
        SendMessage(This.hWndEdit, EM_FORMATRANGE, 0, 0)
        StartDoc(This.PD.hDC, This.DI)
        ResetDC(This.PD.hDC, This.DM)
        StartPage(This.PD.hDC)
        SendMessage(This.hWndEdit, EM_FORMATRANGE, 1, This.FR)
        SendMessage(This.hWndEdit, EM_FORMATRANGE, 1, 0)
        EndPage(This.PD.hDC)
        EndDoc(This.PD.hDC)
        SendMessage(This.hWndEdit, EM_FORMATRANGE, 0, 0)
        DeleteDC(This.PD.hDC)

        This.FR.hdc       = This.hdcMeta
        This.FR.hdcTarget = This.hdcMeta
    End Sub

PUBLIC:
    'This subroutine first checks to see if the defaults were obtained.
    ' Then displays the Page Setup Dialog box (with appropriate units of measure).
    ' If the user clicks "Ok"... the PageSetupDialog structure is updated.
    ' If the user clicks "Cancel" or closes the box... the PageSetupDialog structure remains unchanged.
    ' The accessible .DM structure will be updated by the .StoreMemoryDevMode subroutine.
    Sub PageSetup ()
        If (This.mhDevMode = 0) Then This.GetDevModeDefaults

        If This.UserUnits = 0 Then
            This.PSD.Flags = PSD_MARGINS OR PSD_DISABLEPRINTER Or PSD_INHUNDREDTHSOFMILLIMETERS
        Else
            This.PSD.Flags = PSD_MARGINS OR PSD_DISABLEPRINTER Or PSD_INTHOUSANDTHSOFINCHES
        End If

        'Calling PageSetupDlg with the above flags tells it to display a dialog box
        ' using values that are appropriate for end user.
        PageSetupDlg(This.PSD)

        This.mhDevMode   = This.PSD.hDevMode
        This.mpDevMode = GlobalLock(This.mhDevMode)
        GlobalUnLock(This.mhDevMode)

        This.StoreMemoryDevMode(This.mpDevMode)
    End Sub

    'This subroutine sets the size and location of the preview form to the same size and location as the main application's form.
    ' The size of the "non-previewing" components is based upon the size of the preview form.
    ' The size of the "previewing" components were calculated in the .CalculateDimensions subroutine.
    Sub PrtPreview (ByVal hForm As Long, ByVal hEdit As Long)
        This.hWndForm  = hForm
        This.hWndEdit  = hEdit
        This.PageIndex = 0

        If (This.mhDevMode = 0) Then This.GetDevModeDefaults

        This.CalculateDimensions
        This.CalculatePageCount(hEdit)

        GetWindowRect(hForm, This.WR)

        This.PreviewForm.Left   = This.WR.Left
        This.PreviewForm.Top    = This.WR.Top
        This.PreviewForm.Width  = (This.WR.Right - This.WR.Left)
        This.PreviewForm.Height = (This.WR.Bottom - This.WR.Top)

        This.CtrlPanel.Width   = This.PreviewForm.ClientWidth
        This.CtrlPanel.Height  = 34
        This.CtrlPanel.Caption = "Page: " + STR$(This.PageIndex+1) + " of " + STR$(This.Page.ItemCount - 1)

        This.CtrlFirst.Width   = 32
        This.CtrlFirst.Height  = 32
        This.CtrlFirst.Left    = ((This.CtrlPanel.ClientWidth / 2) - 159)
        This.CtrlFirst.Top     = 2
        This.CtrlFirst.Enabled = False

        This.CtrlPrevious.Width   = 32
        This.CtrlPrevious.Height  = 32
        This.CtrlPrevious.Left    = ((This.CtrlPanel.ClientWidth / 2) - 115)
        This.CtrlPrevious.Top     = 2
        This.CtrlPrevious.Enabled = False

        This.CtrlNext.Width   = 32
        This.CtrlNext.Height  = 32
        This.CtrlNext.Left    = ((This.CtrlPanel.ClientWidth / 2) + 99)
        This.CtrlNext.Top     = 2
        This.CtrlNext.Enabled = IIf(This.Page.ItemCount > 2, True, False)

        This.CtrlLast.Width   = 32
        This.CtrlLast.Height  = 32
        This.CtrlLast.Left    = ((This.CtrlPanel.ClientWidth / 2) + 143)
        This.CtrlLast.Top     = 2
        This.CtrlLast.Enabled = IIf(This.Page.ItemCount > 2, True, False)

        This.CtrlPrint.Width  = 32
        This.CtrlPrint.Height = 32
        This.CtrlPrint.Left   = (This.CtrlPanel.ClientWidth - 60)
        This.CtrlPrint.Top    = 2

        This.ViewScrollbox.Left   = 0
        This.ViewScrollbox.Top    = 35
        This.ViewScrollbox.Width  = This.PreviewForm.ClientWidth
        This.ViewScrollbox.Height = (This.PreviewForm.ClientHeight - 35)

        This.ParentPanel.Left   = This.rcParentSize.Left
        This.ParentPanel.Top    = This.rcParentSize.Top
        This.ParentPanel.Width  = This.rcParentSize.Right
        This.ParentPanel.Height = This.rcParentSize.Bottom

        This.ParentCanvas.Left   = 0
        This.ParentCanvas.Top    = 0
        This.ParentCanvas.Width  = This.ParentPanel.Width
        This.ParentCanvas.Height = This.ParentPanel.Height

        This.ChildPanel.Left   = This.rcChildSize.Left
        This.ChildPanel.Top    = This.rcChildSize.Top
        This.ChildPanel.Width  = This.rcChildSize.Right
        This.ChildPanel.Height = This.rcChildSize.Bottom

        This.ChildCanvas.Left   = 0
        This.ChildCanvas.Top    = 0
        This.ChildCanvas.Width  = This.ChildPanel.Width
        This.ChildCanvas.Height = This.ChildPanel.Height

        This.hdcPanel   = GetDC(This.ChildPanel.Handle)

        This.GeneratePage(This.hWndEdit, This.PageIndex)

        This.PreviewForm.Show
    End Sub

    'And last... but not least... this subroutine displays the Print Dialog box
    ' and prints the document according to user selected options.
    'Special Note; If the selected printer does NOT support
    ' multiple copies and collate... you'll have to write your
    ' own coding to handle these situations. If it does... you
    ' do not have to modify the code below. That information
    ' is stored in the .SHD file that's automatically generated.
    ' (the .SPL file only contains an image (or images) of the printed page (or pages))
    Sub PrtDialog (ByVal hForm As Long, ByVal hEdit As Long)
        Dim iTextOut      As Long
        Dim iTextAmt      As Long
        Dim szString      As String * 200

        If (This.mhDevMode = 0) Then This.GetDevModeDefaults

        This.CalculateDimensions
        This.CalculatePageCount(hEdit)

        MemSet(This.PD, 0, SizeOf(This.PD))
        This.PD.lStructSize = SizeOf(This.PD)
        This.PD.hInstance   = Application.hInstance
        This.PD.hWndOwner   = hForm
        This.PD.hDevMode    = This.PSD.hDevMode
        This.PD.hDevNames   = This.PSD.hDevNames
        This.PD.Flags       = PD_ALLPAGES Or PD_HIDEPRINTTOFILE Or PD_NOSELECTION Or PD_RETURNDC Or PD_USEDEVMODECOPIESANDCOLLATE
        This.PD.nFromPage   = 1
        This.PD.nToPage     = (This.Page.ItemCount - 1)
        This.PD.nMinPage    = 1
        This.PD.nMaxPage    = (This.Page.ItemCount - 1)
        This.PD.nCopies     = 1

        'Calling PrintDlg with the above flags tells it to display a dialog box
        ' using the values assigned to the PD structure fields.
        If PrintDlg(This.PD) Then
            MemSet(This.FR, 0, SizeOf(This.FR))
            This.FR.hdc           = This.PD.hDC
            This.FR.hdcTarget     = This.PD.hDC
            This.FR.chrg_cpMin    = Val(This.Page.Item(This.PD.nFromPage - 1))
            This.FR.chrg_cpMax    = Val(This.Page.Item(This.PD.nToPage)) - 1

            This.FR.rcPage_Left   = This.rcPrint.Left
            This.FR.rcPage_Top    = This.rcPrint.Top
            This.FR.rcPage_Right  = This.rcPrint.Right
            This.FR.rcPage_Bottom = This.rcPrint.Bottom

            This.FR.rc_Left       = This.FR.rcPage_Left
            This.FR.rc_Top        = This.FR.rcPage_Top
            This.FR.rc_Right      = This.FR.rcPage_Right
            This.FR.rc_Bottom     = This.FR.rcPage_Bottom

            If This.DocName = "" Then
                szString = "QPrint: NoName"
            Else
                szString = "QPrint: " & Mid$(This.DocName, rInstr(This.DocName, "\") + 1)
            End If

            MemSet(This.DI, 0, SizeOf(This.DI))
            This.DI.cbSize      = SizeOf(This.DI)
            This.DI.lpszDocName = VarPtr(szString)
            This.DI.lpszOutput  = 0

            SendMessage(hEdit, EM_FORMATRANGE, 0, 0)
            StartDoc(This.PD.hDC, This.DI)
            StartPage(This.PD.hDC)
            iTextOut = 0
            iTextAmt = This.FR.chrg_cpMax

            While (iTextOut < iTextAmt)
                iTextOut = SendMessageAPI(hEdit, EM_FORMATRANGE, 1, This.FR)
                If (iTextOut < iTextAmt) Then
                    EndPage(This.PD.hDC)
                    StartPage(This.PD.hDC)
                    This.FR.chrg_cpMin = iTextOut
                End If
            Wend

            SendMessage(hEdit, EM_FORMATRANGE, 1, 0)
            EndPage(This.PD.hDC)
            EndDoc(This.PD.hDC)
            SendMessage(hEdit, EM_FORMATRANGE, 0, 0)

            DeleteDC(This.PD.hDC)
        End If
    End Sub

    Constructor
        'Set up the Preview form's and it's components' properties and events
        ' (the size and location of each will be determined in the .PrtPreview subroutine)
        PreviewForm.BorderStyle    = bsDialog
        PreviewForm.Caption        = " Print Preview"
        PreviewForm.OnShow         = This.PreviewForm_OnShow
        PreviewForm.OnClose        = This.PreviewForm_OnClose

        CtrlPanel.Parent           = This.PreviewForm
        CtrlPanel.BevelOuter       = bvNone
        CtrlPanel.BevelInner       = bvNone
        CtrlPanel.BevelWidth       = 0

        CtrlFirst.Parent           = This.CtrlPanel
        CtrlFirst.Layout           = blBMPBottom
        CtrlFirst.Font.Name        = "Webdings"
        CtrlFirst.Font.Size        = 14
        CtrlFirst.Caption          = Chr$(57) '"9"
        CtrlFirst.OnClick          = This.CtrlFirst_OnClick

        CtrlPrevious.Parent        = This.CtrlPanel
        CtrlPrevious.Font.Name     = "Webdings"
        CtrlPrevious.Font.Size     = 14
        CtrlPrevious.Caption       = Chr$(51) '"3"
        CtrlPrevious.OnClick       = This.CtrlPrevious_OnClick

        CtrlNext.Parent            = This.CtrlPanel
        CtrlNext.Font.Name         = "Webdings"
        CtrlNext.Font.Size         = 14
        CtrlNext.Caption           = Chr$(52) '"4"
        CtrlNext.OnClick           = This.CtrlNext_OnClick

        CtrlLast.Parent            = This.CtrlPanel
        CtrlLast.Layout            = blBMPBottom
        CtrlLast.Font.Name         = "Webdings"
        CtrlLast.Font.Size         = 14
        CtrlLast.Caption           = Chr$(58) '":"
        CtrlLast.OnClick           = This.CtrlLast_OnClick

        CtrlPrint.Parent           = This.CtrlPanel
        CtrlPrint.Font.Name        = "Webdings"
        CtrlPrint.Font.Size        = 20
        CtrlPrint.Caption          = Chr$(202) '""
        CtrlPrint.OnClick          = This.CtrlPrint_OnClick

        ViewScrollbox.Parent       = This.PreviewForm
        ViewScrollBox.HorzTracking = True
        ViewScrollBox.VertTracking = True
        ViewScrollbox.Color        = RGB(000,000,000)

        ParentPanel.Parent         = This.ViewScrollbox
        ParentPanel.BevelOuter     = bvNone
        ParentPanel.BevelInner     = bvNone
        ParentPanel.BevelWidth     = 0
        ParentPanel.Color          = RGB(255,255,255)

        ParentCanvas.Parent          = This.ParentPanel
        ParentCanvas.OnPaint         = This.ParentCanvas_OnPaint

        ChildPanel.Parent            = This.ParentPanel
        ChildPanel.BevelOuter        = bvNone
        ChildPanel.BevelInner        = bvNone
        ChildPanel.BevelWidth        = 0
        ChildPanel.Color             = RGB(255,255,255)

        ChildCanvas.Parent          = This.ChildPanel
        ChildCanvas.OnPaint         = This.ChildCanvas_OnPaint

        'Set the (global) UserUnits variable during initialization
        UserUnits = This.GetLocaleInfo_iMeasure()
    End Constructor
End Type

'Just some trivia information
'inch(s) * 25.4 = mm(s)
'inch(s) * 254  = cm(s)
'  mm(s) / 25.4 = inch(s)
'  cm(s) / 254  = inch(s)
'mm(s) * 0.03936996 = inch(s) (round to however many decimal places you need)
'1440 twips per inch
'56.692913385826771653543307086614 twips per mm
'5.6692913385826771653543307086614 twips per cm
'screen DPIs are device dependent (on my computer they are 96)
'printer DPIs are device dependent (on my computer they are 600)
'(DPI = Dots Per Inch (both x horizontal and y vertical))
